@@ -1,20 +1,15 @@
+import itertools
 import time
-from copy import deepcopy
 from itertools import chain, combinations
 
 import numpy as np
 import pandas
 
 pandas.options.mode.chained_assignment = None
-from rpy2.robjects import pandas2ri
-from rpy2.robjects.packages import importr
-from .accessory import pdag2dag, to_bnlearn, g_test, zf_test, pdag2cpdag, find_causes, pairwise, dag2cpdag
+from .accessory import pdag2dag, to_bnlearn, g_test, z_test, find_causes, pairwise, dag2cpdag, \
+    independence_test, orient_v_structure, orient_propagation
 from .score import local_score
 
-pandas2ri.activate()
-
-base, bnlearn = importr('base'), importr('bnlearn')
-
 
 def pre_process(data):
     '''
@@ -42,7 +37,7 @@ def post_process(graph, dc):
     :return: expanded graph
     '''
     for node in dc:
-        graph = bnlearn.add_node(graph, node)
+        graph += '[' + node + ']'
     return graph
 
 
@@ -129,7 +124,7 @@ def hc(data, method='standard', score_function='default', debug=False):
         time_preprocess = time.time() - start_time
 
         diff = 1
-        cache_dag = [deepcopy(dag)]
+        cache_dag = [{var: {'par': [v for v in dag[var]['par']], 'nei': [v for v in dag[var]['nei']]} for var in dag}]
         while diff > 0:
             diff = 0
             edge_candidate = []
@@ -281,7 +276,7 @@ def hc(data, method='standard', score_function='default', debug=False):
                     else:
                         dag[vi]['par'] = sorted(dag[vi]['par'] + [par_vi])
             dag, candidate = dag_update(dag, edge_candidate, candidate, diff, debug)
-            cache_dag.append(deepcopy(dag))
+            cache_dag.append({var: {'par': [v for v in dag[var]['par']], 'nei': [v for v in dag[var]['nei']]} for var in dag})
         dag = to_bnlearn(dag)
         time_total = time.time() - time_total
         if debug:
@@ -384,28 +379,79 @@ def hc(data, method='standard', score_function='default', debug=False):
             print('check cycle cost: %.2f%%' % (time_checkcycle / time_total * 100))
             print('compute score cost: %.2f%%' % (time_score / time_total * 100))
             print('others cost: %.2f%%' % (
-                        (time_total - time_preprocess - time_checkcycle - time_score) / time_total * 100))
-    elif method == 'bnlearn':
-        dag = bnlearn.hc(data, score=score_function, debug=debug)
+                    (time_total - time_preprocess - time_checkcycle - time_score) / time_total * 100))
     else:
         raise Exception('The input method:' + method + 'is invalid.')
     return post_process(dag, dc)
 
 
-def pc_stable(data, test='default', debug=False):
+def pc_stable(data, test='default', threshold=0.05, debug=False):
+    '''
+
+    Args:
+        data: original input data (in pandas dataframe format)
+        test: statistical CI test (mutual information for categorical data and fisher's z test for continuous data)
+        threshold: threshold for p-value
+        debug: whether or not to output debug information
+
+    Returns:
+
+    '''
     if test == 'default':
-        test = 'mi' if all(data[var].dtype.name == 'category' for var in data) else 'zf'
+        if all(data[var].dtype.name == 'category' for var in data):
+            test = 'g-test'
+        elif all(data[var].dtype.name != 'category' for var in data):
+            test = 'z-test'
+        else:
+            raise Exception('Mixed type of data is not supported yet.')
     data, dc = pre_process(data)
-    learned = bnlearn.pc_stable(data, test=test, debug=debug)
-    return post_process(learned, dc)
-
-
-def h2pc(data, debug=False):
-    data, dc = pre_process(data)
-    dag = bnlearn.h2pc(data, debug=debug)
+    varnames = data.columns.tolist()
+    if all(data[var].dtype.name == 'category' for var in data):
+        data = data.apply(lambda x: x.cat.codes).to_numpy()
+    elif all(data[var].dtype.name != 'category' for var in data):
+        data = data.to_numpy()
+    else:
+        raise Exception('Mixed data is not supported.')
+    # initialise pc set as full for all variables
+    pc = {var: [v for v in varnames if v != var] for var in varnames}
+    sepset = {var: {} for var in varnames}
+    l = 0
+    # find PC set for each variable
+    while max({key: len(value) for key, value in pc.items()}.values()) > l:
+        pc_copy = {var: [v for v in pc[var]] for var in pc}
+        for var in varnames:
+            for adj in pc_copy[var]:
+                if adj not in sepset[var]:
+                    pc_var = [v for v in pc_copy[var]]
+                    pc_var.remove(adj)
+                    for con in itertools.combinations(pc_var, l):
+                        cols = np.array([varnames.index(v) for v in [var, adj] + list(con)])
+                        test_result, p = independence_test(data, cols, test=test, varnames=varnames,
+                                                           threshold=threshold)
+                        if test_result:
+                            if debug:
+                                print(var + ' and ' + adj + ' are independent given ' + con + ', p value=' + p)
+                            pc[var].remove(adj)
+                            pc[adj].remove(var)
+                            sepset[var][adj] = con
+                            sepset[adj][var] = con
+                            break
+        l += 1
+    # orient v-structrues
+    cpdag = orient_v_structure(data, varnames, pc, sepset, test, threshold=threshold)
+    # orientation propagation
+    cpdag = orient_propagation(cpdag)
+    dag = pdag2dag(cpdag)
+    dag = to_bnlearn(dag)
     return post_process(dag, dc)
 
 
+# def h2pc(data, debug=False):
+#     data, dc = pre_process(data)
+#     dag = bnlearn.h2pc(data, debug=debug)
+#     return post_process(dag, dc)
+
+
 def ges(data, method='complete', score_function='default', prune=True, thres=0.1, debug=False):
     '''
     greedy equivalence search algorithm
@@ -458,7 +504,7 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
             for i in range(len(varnames) - 1):
                 for j in range(i + 1, len(varnames)):
                     cols = [i, j]
-                    p = g_test(data, cols) if datatype == 'category' else zf_test(data, cols)
+                    p = g_test(data, cols) if datatype == 'category' else z_test(data, varnames, cols)
                     if p > thres:
                         candidate[varnames[i]].remove(varnames[j])
                         candidate[varnames[j]].remove(varnames[i])
@@ -502,7 +548,7 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                     cpdag[operator_candidate[1]]['par'].append(t)
                     cpdag[operator_candidate[1]]['nei'].remove(t)
                     cpdag[t]['nei'].remove(operator_candidate[1])
-                cpdag = pdag2cpdag(cpdag)
+                cpdag = dag2cpdag(cpdag)
             else:
                 break
         # backward phase
@@ -554,7 +600,7 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                         cpdag[h]['nei'].remove(operator_candidate[0])
                         cpdag[operator_candidate[0]]['nei'].remove(h)
                         cpdag[h]['par'].append(operator_candidate[0])
-                cpdag = pdag2cpdag(cpdag)
+                cpdag = dag2cpdag(cpdag)
             else:
                 break
     elif method in ['pw', 'ipw', 'aipw']:
@@ -584,8 +630,8 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                                                             cache_data, cache_weights, method)
                     if len(cache_data[W]) != 0:
                         cols = [i, j]
-                        p = g_test(cache_data[W], cols, cache_weights[W]) if datatype == 'category' else zf_test(
-                            cache_data[W], cols, cache_weights[W])
+                        p = g_test(cache_data[W], cols, cache_weights[W]) if datatype == 'category' else z_test(
+                            cache_data[W], varnames, cols)
                         if p > thres:
                             candidate[varnames[i]].remove(varnames[j])
                             candidate[varnames[j]].remove(varnames[i])
@@ -635,7 +681,7 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                     cpdag[operator_candidate[1]]['par'].append(t)
                     cpdag[operator_candidate[1]]['nei'].remove(t)
                     cpdag[t]['nei'].remove(operator_candidate[1])
-                cpdag = pdag2cpdag(cpdag)
+                cpdag = dag2cpdag(cpdag)
             else:
                 break
         # backward phase
@@ -694,7 +740,7 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                         cpdag[h]['nei'].remove(operator_candidate[0])
                         cpdag[operator_candidate[0]]['nei'].remove(h)
                         cpdag[h]['par'].append(operator_candidate[0])
-                cpdag = pdag2cpdag(cpdag)
+                cpdag = dag2cpdag(cpdag)
             else:
                 break
     time_total = time.time() - time_total
@@ -704,15 +750,9 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
 
     dag = to_bnlearn(pdag2dag(cpdag))
     if len(dc):
-        modelstring = str(bnlearn.modelstring(dag))
-        modelstring = modelstring[5:len(modelstring) - 2]
         for node in dc:
-            modelstring += '[' + node + ']'
-        cpdag = bnlearn.cpdag(bnlearn.model2network(modelstring))
-    else:
-        cpdag = bnlearn.cpdag(dag)
-
-    return cpdag
+            dag += '[' + node + ']'
+    return dag
 
 
 def all_subsets(ss):
